import configparser
import importlib
import os
import os.path    as osp
import sqlite3    as sq
import warnings
import yaml       as ya
from   io         import StringIO
from   pandas     import concat, DataFrame, Series, read_sql_query, read_csv
from   sqlite3    import Error

###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from .model_checker import check_model_terms
    from .model_coder   import model_to_code

except:
    from model_checker import check_model_terms
    from model_coder   import model_to_code
    
###############################################################################
#Globals
###############################################################################
MBase  = None
UBase  = None
_dir   = osp.dirname(osp.abspath(__file__))
userid = 'usr'

table_sql = '''
CREATE TABLE IF NOT EXISTS "models" (
    "id"           TEXT,
	"system_type"  TEXT,
	"states"	   TEXT,
    "parameters"   TEXT,
    "inputs"       TEXT,
    "equations"    TEXT,
    "ia"           TEXT,
    "descriptions" TEXT,
    "active"       INTEGER DEFAULT 1,
    UNIQUE(system_type)
);
'''

###############################################################################
#Database and Table Construction
###############################################################################
def create_connection(db_file):
    '''
    Creates database specified by db_file.
    '''
    db = sq.connect(db_file)
    return db

def create_table(db, *args):
    '''
    Creates table.
    Use args to add additional sql commands.
    '''
    try:
        c = db.cursor()

        for arg in args:
            c.execute(arg)
    except Error as e:
        raise e

###############################################################################
#Constructor
###############################################################################
def make_core_model(system_type, states, parameters, inputs, equations, descriptions=None, ia='', **kwargs):
    '''Returns a dictionary with the keys id, system_type, states, parameters, inputs, equations, ia.
    Otherwise referred to as a core_model in the BMSS2 documentation. Redundant
    states, parameters and inputs are disallowed.
    
    :param system_type: A string of keywords serving as a unique identifier for 
        the core_model separated by commas, will be formatted so there is one space
        after each comma, keywords should be in CamelCase
    :type system_type: str
    :param states: A list of strings corresponding to state names used in the core_model
    :type states: list
    :param parameters: A list of strings corresponding to parameter names used in the core_model
    :type parameters: list
    :param inputs: A list of strings corresponding to input names used in the core_model
    :type inputs: list
    :param equations: A list of strings corresponding to lines of equations used in the core_model
        where the lines form coherent Python code when joined by '\n'.join
    :type equations: list
    :param descriptions: A description of the model.
    :type descriptions: dict, optional
    :param ia: For IA results as a string that can be read into csv format. Avoid using this argument.
    :type ia: string, optional
    :kwargs: Will be ignored
    '''    
    if type(system_type) == str:
        system_type1 = ', '.join([s.strip() for s in system_type.split(',')])
    else:
        system_type1 = ', '.join(system_type)
        return make_core_model(system_type1, states, parameters, inputs, equations, descriptions, ia='')
    
    states1       = list(states)
    parameters1   = list(parameters)
    inputs1       = list(inputs)
    equations1    = list(equations)
    descriptions1 = descriptions if descriptions else {}

    core_model = {'id'           : '',
                  'system_type'  : system_type1,
                  'states'       : states1,
                  'parameters'   : parameters1,
                  'inputs'       : inputs1,
                  'equations'    : equations1, 
                  'ia'           : ia,
                  'descriptions' : descriptions1,
                  }
    
    for key in ['states', 'parameters', 'inputs', 'equations']:
        if not all([type(x)==str for x in core_model[key]]):
            raise Exception('Invalid name in ' + key + '. Only strings are allowed.')
    
    if type(core_model['ia']) != str:
        raise Exception('Invalid name in ia. Only strings are allowed.')
    
    is_valid, text = check_model_terms(core_model)
    if not is_valid:
        warnings.warn('Error in ' + str(system_type1) + ' when checking terms: ' + text)
        # raise Exception('Error in ' + str(system_type1) + ' when checking terms: ' + text)
    
    return core_model

def copy_core_model(core_model):
    keys = ['id', 'system_type', 'states' , 'parameters', 'inputs', 'equations', 'ia', 'descriptions']
    
    new_core_model = {}
    for key in keys:
        try:
            value = core_model[key].copy()
        except:
            value = core_model[key]
    
        new_core_model[key] = value
    return new_core_model
        
###############################################################################
#Model Storage
###############################################################################
def add_to_database(core_model, dialog=True):
    '''Accepts a core_model and adds it to UBase.
    '''
    if 'BMSS' in core_model['system_type']:
        raise Exception('system_type cannot contain "BMSS" as keyword.')
    return backend_add_to_database(core_model, database=UBase, dialog=dialog)

###############################################################################
#Supporting Functions
###############################################################################
def backend_add_to_database(core_model, database, dialog=False):
    '''Supporting function for add_to_database. Do not run.
    
    :meta private:
    '''
    global MBase
    global UBase
    global userid
     
    system_type    = core_model['system_type']
    make_new_id    = True
    existing_model = quick_search(system_type, error_if_no_result=False, active_only=False)
    d              = 'Mbase' if database == MBase else 'UBase'
    if existing_model:
        if dialog:
            while True:
                x = input('Overwrite existing model? (y/n): ')
                if x.lower() == 'y':
                    break
                elif x.lower() == 'n':
                    return existing_model['id']
                else:
                    continue
                    
        if system_type == existing_model['system_type']:
            core_model['id'] = existing_model['id']
            make_new_id = False
                

    row    = string_dict_values(core_model)   
    row_id = add_row('models', row, database)

    #Update id based on row number if the model is new
    if make_new_id:
        model_id = 'bmss' + str(row_id) if database == MBase else userid + str(row_id)

        update_value_by_rowid(row_id, 'id', model_id, database)
    
    else:
        model_id = core_model['id']
    
    o = 'Added model ' if make_new_id else 'Modified model '
    n =  model_id      if make_new_id else core_model['id']
    print(o + n + ' to '+ d)
    
    model_to_code(core_model, local=False)
    return model_id

def string_dict_values(core_model):
    '''Converts core_model to string
    
    :meta private:
    '''
    model_dict = {key: str(core_model[key]) for key in core_model}
    
    return model_dict

def update_value_by_rowid(row_id, column_id, value, database):
    '''Supporting function for backend_add_to_database. Do not run.
    
    :meta private:
    '''
    with database as db:
        comm = "UPDATE models SET " + column_id + " = '" + str(value) + "' WHERE rowid = " + str(row_id) 
        cur  = db.cursor()
        cur.execute(comm)
    
def add_row(table, row, database):
    '''Supporting function for backend_add_to_database. Do not run.
    
    :meta private:
    '''
    row_   = '(' + ', '.join([k for k in row.keys()]) + ', active)'
    values = tuple(row.values()) + ('1',)
    
    with database as db:
        comm = 'REPLACE INTO ' + table + str(row_) + ' VALUES(' + ','.join(['?']*len(values))+ ')'
        cur  = db.cursor()
        cur.execute(comm, values)

    return cur.lastrowid

###############################################################################
#Search
###############################################################################
def search_database(keyword, search_type='system_type', database=None, active_only=True):
    '''Searches database for core_model data structure based on a keyword and a 
    key(field) in the core_model. Returns a list of core_model dictionaries.
    
    :param keyword: A string that will be matched against entries in the database
    :type keyword: str
    :param search_type: A string corresponding to any key in the core_model, will 
        be used for matching
    :type search_type: str
    :param database: Can be either MBase or UBase, if None, this function will search
        both databases, defaults to None
    :type database: SQL Connection, optional 
    :param active_only: A boolean for backend use, if True, this limits the search 
        to rows where the value of active is True, defaults to True
    :type active_only: bool, optional
    '''
    global MBase
    global UBase
    
    keyword1  = keyword if type(keyword) == str else ', '.join(keyword)
    comm      = 'SELECT id, system_type, states, parameters, inputs, equations, ia, descriptions FROM models WHERE ' + search_type + ' LIKE "%' + keyword1 
    result    = []
    if active_only:
        comm      += '%" AND active = 1;'
    else:
        comm += '%";'

    databases = [database] if database else [MBase, UBase]
    for database in databases:
        with database as db:
            cursor    = db.execute(comm)
            models    = cursor.fetchall()    
        
        columns = database.execute('PRAGMA table_info(models);')
        columns = columns.fetchall()
        columns = [column[1] for column in columns if column[1]!='active']
        models  = [dict(zip(columns, model)) for model in models]
        models = [process_model(model) for model in models]
        result += models
    
    return result

def process_model(model):
    '''
    :meta private:
    '''
    
    result = {}
    for key, value in model.items():
        if key in ['id', 'system_type', 'ia']:
            result[key] = value
        else:
            try:
                result[key] = eval(value)
            except Exception as e:
                system_type = model['system_type']
                msg         = 'An error occurred when calling eval on {} for {}'.format(key, system_type)
                raise Exception(msg, *e.args)
    return result

def quick_search(system_type, error_if_no_result=True, **kwargs):
    '''Searches both system_type/id and returns an exact match in system_type/id.
    Raises an error when no matches are found if error_if_no_result is set to True.
    '''
    core_models = search_database(system_type, search_type='system_type', **kwargs)
    for core_model in core_models:
        if core_model['system_type'] == system_type:
            return core_model
    if error_if_no_result:
        raise Exception('Could not retrieve model with system_type ' + str(system_type))
        
    else:
        return

def list_models(database=None):
    '''Returns a list of system_types
    '''
    global MBase
    global UBase
        
    if database:
        with database as db:
            comm      = 'SELECT system_type FROM models WHERE active = 1;'
            cursor    = db.execute(comm)
            models    = [m[0] for m in cursor.fetchall()]
        
        return models
    else:
        return list_models(MBase) + list_models(UBase)

def get_model_function(system_type, local=False):
    '''Supporting function for get_model_function. Do not run.
    
    :meta private:
    '''
    model_name     = system_type.replace(', ', '_')
    if local:
        module = importlib.import_module(model_name)
    else:
        module = importlib.import_module('.model_functions.' + model_name, 'BMSS.models')
    model_function = getattr(module, 'model_'+  model_name )
    
    return model_function

###############################################################################
#Interfacing with Pandas
###############################################################################
def to_df(database=None):
    '''Returns a copy of the databases as a pandas DataFrame.
    
    :param database: The database to be read, if None, both databases will be read,
        defaults to None
    :type database: SQL Connection
    '''
    if database:
        with database as db:
            df = read_sql_query("SELECT * from models", db)
        return df
    else:
        global MBase
        global UBase
        df = to_df(MBase), to_df(UBase)
        df = concat(df, ignore_index=True)
        return df

def backend_from_df(df, database):
    '''For backend maintenance only.
    
    :meta private:
    '''
    with database as db:
        return df.to_sql('models', db, if_exists='replace', index=False)

###############################################################################
#Interfacing with Configparser
###############################################################################
def from_config(filename):
    '''Reads a file and returns a core_model data structure.
    
    :param filename: The name of the file to read.
    :type filename: str
    '''
    config = configparser.ConfigParser()
    config.optionxform  = lambda option: option
    model  = {'system_type'  : [],
              'states'       : [], 
              'parameters'   : [],
              'inputs'       : [],
              'equations'    : [],
              'ia'           : '',
              'descriptions' : {}
              }
    with open(filename, 'r') as file:
        config.read_file(file)
        
    for key in config.sections():
        if key not in model:
            continue
        if key == 'ia':
            line = config[key][key].strip()
        elif key == 'equations':
            line = config[key][key].replace('\n', ',').split(',')
            line = [s.strip() if s else '' for s in line]
            line = line if line[0] else line[1:]
        elif key == 'descriptions':
            temp = config[key]
            line = {k: temp[k] for k in temp}
        else:
            line = config[key][key].replace('\n', ',').split(',')
            line = [s.strip() for s in line if s]
        
        model[key] = line
    
    return make_core_model(**model)

def to_config(core_model, filename):
    '''Exports a core_model data structure to a config file.
    
    :param core_model: The core_model to be exported
    :type core_model: dict
    :param filename: The name of the file to write to
    :type filename: str 
    '''
    config = configparser.ConfigParser()
    
    for key in core_model:
        if not core_model[key]:
            continue
        if type(core_model[key]) == str:
            line = core_model[key]
        elif type(core_model[key]) == dict:
            config[key] = core_model[key]
            continue
        elif key == 'equations':
            line = '\n' + '\n'.join(core_model[key])
        else:
            line = ', '.join(core_model[key])
        config[key] = {key:line}
        
    with open(filename, 'w') as configfile:
        config.write(configfile)
    
    return config

###############################################################################
#Direct Config to Database
###############################################################################
def config_to_database(filename, dialog=True):
    '''Reads a config file containing information for a core_model data structure 
    and adds the core_model to the database.
    
    :param filename: Name of the file to be read
    :type filename: str
    :param dialog: For backend use, defaults to True
    :type dialog: bool, optional
    '''
    global UBase
    return backend_config_to_database(filename, database=UBase, dialog=dialog)

def backend_config_to_database(filename, database, dialog=False):
    '''For backend maintenance. Do not run.
    
    :meta private:
    '''
    core_model = from_config(filename)
    backend_add_to_database(core_model, database, dialog=dialog)
    return core_model['system_type']

###############################################################################
#Updateing IA
###############################################################################
def update_ia(core_model, new_row, save=True):
    '''
    Appends new_row to the core_model['ia'] where new_row can be a dict or Series.
    
    If save is True and the core_model is in the database, the changes will be 
    applied to database.
    '''
    global MBase
    global UBase
    
    s  = new_row if type(new_row) in [Series, DataFrame] else Series(new_row) 
        
    if core_model['ia']:
        df = read_ia(core_model)
        
        try:
            s = s[df.columns]
        except:
            raise Exception('Mismatch in column names. \nExpected: ' +  
                            str(df.columns) + '\nReceived: ' + str(s.index))
            
        #Check if inputs are duplicated
        try:
            row_num          = df['input'][s['input'] == df['input']].index[0]
            df.iloc[row_num] = s
            print('Updated row ' + str(row_num) + ' in ' + str(core_model['system_type']))
        except:
            df = df.append(s, ignore_index=True)
            print('Added row to ' + str(core_model['system_type']))
    
    else:
        df = DataFrame(s).T
    
    ia_string        = df.to_csv(index=False)
    core_model['ia'] = ia_string
    
    if save:
        system_type = core_model['system_type']
        
        if quick_search(system_type, error_if_no_result=False):
            _backend_modify_database(system_type, {'ia': ia_string})
        else:
            print('Changes to ' + str(system_type) + ' not saved to database.')
    else:
        print('Changes to ' + str(core_model['system_type']) + ' not saved to database.')
    return ia_string

def read_ia(core_model, mode='df', filename=''):
    '''
    Reads the string in core_model['ia'] and converts into a DataFrame.
    
    If mode is df, the DataFrame is returned.
    
    If mode is csv, the string is written to a csv file.
    
    If mode is yaml, the method to_dict("record") is called and the resulting
    dict is converted to .yaml.
    '''
    df = read_csv(StringIO(core_model['ia']), header=[0, 1])    
    
    if mode == 'csv':
        return df.to_csv(filename, index=False)
    elif mode == 'yaml':
        d = df.to_dict('record')
        return ya.dump(d, filename=filename)
    else:
        return df

def reset_ia(system_type):
    if quick_search(system_type, error_if_no_result=False):
        _backend_modify_database(system_type, {'ia': ''})
 
###############################################################################
#Modification
###############################################################################
def _backend_modify_database(system_type, data):
    '''
    Updates the row containing system_type with data.
    Automatically determines if the system_type is part of MBase or UBase.
    Not meant to be used on unsanitized data.
    '''
            
    if 'BMSS' in system_type:
        database = MBase
    else:
        database = UBase
    
    with database as db:
        for column, value in data.items():
            comm = "UPDATE models SET " + str(column) + " = '" + str(value) + "' WHERE system_type = '" + str(system_type) + "'"
            cur  = db.cursor()
            cur.execute(comm)
            print('Changes to ' + str(system_type) + ' saved to database.')
        
    return

###############################################################################
#Deletion
###############################################################################
def delete(system_type):
    '''Sets the value of the active column of the core_model in the .db file to 0.
    Can be restored using the restore function.
    '''
    database = UBase
    
    with database as db:
        comm = 'UPDATE models SET active = 0 WHERE system_type = "' + system_type + '";'
        cur  = db.cursor()
        cur.execute(comm)

def restore(system_type):
    '''Sets the value of the active column of the core_model in the .db file to 1.
    '''
    database = UBase

    with database as db:
        comm = 'UPDATE models SET active = 1 WHERE system_type = "' + system_type + '";'
        cur  = db.cursor()
        cur.execute(comm)

def true_delete(system_type, database):
    if not quick_search(system_type, active_only=False, error_if_no_result=False, database=database):
        print(str(system_type) + ' could not be deleted as it was not found.')

    with database as db:
        comm = 'DELETE FROM models WHERE system_type="'+ system_type + '";'    
        db.execute(comm)
        print('Removed ' + system_type)

###############################################################################
#Function for Setup
###############################################################################
def setup():
    global UBase
    global MBase
    global userid
    global _dir
    
    #Database settings
    config = configparser.ConfigParser()
    with open(osp.join(_dir, 'database_settings.ini'), 'r') as file:
        config.read_file(file)
    
    userid = config['userid']['userid']
    
    #Connect to dabatases
    for filename in ['MBase.db', 'UBase.db']:
        db_file     = osp.join(_dir, filename)#osp.join(osp.realpath(osp.split(__file__)[0]), filename)
        database    = create_connection(db_file)
        cursor      = database.execute("SELECT name FROM sqlite_master WHERE type='table';")
        table_names = [table[0] for table in cursor]
        if 'models' not in table_names:
            create_table(database, table_sql)
        
        if filename == 'MBase.db':
            MBase = database
        else:
            UBase = database
    
    #Check required folders are present
    lst = os.listdir(_dir)
    for folder in ['model_functions', 'ia_results', 'BMSS_markup']:
        if folder not in lst:
            os.mkdir(osp.join(_dir, folder))
    print('Connected to MBase_models, UBase_models')
    
    return database

def change_userid(new_userid):
    global userid
    
###############################################################################
#Initialization
###############################################################################
setup()

if __name__ == '__main__':
    from   pathlib import Path
    import os
    
    # #For loading markup files in markup to MBase
    # markup_directory = Path(os.getcwd()) /'markup'
    # for f in os.listdir(markup_directory):
    #     filename = markup_directory/f
    #     backend_config_to_database(filename, MBase)
    
    __model__ = {'system_type' : ['DUMMY', 'DUMMY'],
                  'states'      : ['mRNA', 'Pep'], 
                  'parameters'  : ['syn_mRNA', 'deg_mRNA', 'syn_Pep', 'deg_Pep', 'Ki'],
                  'inputs'      : ['Ind'],
                  'equations'   : ['dmRNA = syn_mRNA*Ind/(Ind + Ki) - deg_mRNA*mRNA',
                                  'dPep  = syn_Pep*mRNA - deg_Pep'
                                  ],
                  'ia'          : ''
                 
                  }
    
    # core_model = make_core_model(**__model__)
    # add_to_database(core_model)
    # search_result = quick_search('DUMMY, DUMMY')
    # print(search_result)
    
    # filename = 'dummy.ini'
    # core_model = from_config(filename)
    # to_config(core_model, 'writeout.ini')
    


import importlib
import re
from   itertools import product
from   pandas    import DataFrame
from   textwrap  import dedent

###############################################################################
#Globals
###############################################################################
mod_prefix = 'BMSS'

###############################################################################
#Supporting Functions for Import
###############################################################################
def split_line(line, delimiter=','):
    result = []
    for x in line.split(delimiter):
        x1 = x.strip()
        if len(x1) > 0:
            result.append(x1)
    return result

def quote(line, quote_char=1):
    q = "'" if quote_char == 1 else '"'
    return [q+x+q for x in line]

###############################################################################
#Import
###############################################################################
def import_txt(filename):
    with open(filename+'.txt', 'r') as file:
        return load_model(file)

def load_model(filestream):
    content_type = ''
    content = {'states'     : {}, 
               'parameters' : [], 
               'inputs'     : {},
               'ics'        : {}, 
               'equations'  : [],
               'known'      : {}, 
               'measured'   : [],
               'decomp'     : []
               }
    filestream1 = filestream.split('\n') if type(filestream) == str else filestream
    for line in filestream1:
        if len(line.rstrip()) == 0:
            continue
        elif line[0] == '%':
            continue
        elif line[0] == '#':
            content_type = line[1:].strip()
        
        elif content_type == 'equations':
            content['equations'].append(line.strip())
        elif content_type in ['inputs', 'known', 'states']:
            line1 = [s.strip() for s in line.strip().split(',')]
            line1 = [s.split('=') for s in line1 if s]
            for pair in line1:
                key, value                 = pair[0].strip(), pair[1].strip()
                content[content_type][key] = eval(value)

        elif content_type == 'decomp':
            line1 = [s.strip() for s in line.strip().replace(']', '[').split('[')]
            line1 = [s.strip() for s in line1 if s and s.strip() != ',']
            line1 = [[ss.strip() for ss in s.split(',')] for s in line1]
            content[content_type].extend(line1)
        else:
            line1 = [s.strip() for s in line.strip().split(',')]
            line1 = [s for s in line1 if s]
            content[content_type].append(line1)
    
    content['ics']    = content['states'].copy()
    content['states'] = [list(content['states'].keys())]
    
    states             = [state for lst in content['states']     for state in lst]
    measured           = [state for lst in content['measured']   for state in lst]
    parameters         = [param for lst in content['parameters'] for param in lst]
    input_conditions   = content['inputs']
    measured_states    = [state for state in states if state in measured]
    unknown_parameters = [param for param in parameters if param not in content['known']]
    
    
    data = {'variables'          : content['states'] + content['parameters'] + list(content['inputs'].keys()),
            'equations'          : content['equations'],
            'measured_states'    : measured_states,
            'states'             : states,
            'unknown_parameters' : unknown_parameters,
            'input_conditions'   : input_conditions,
            'ics'                : content['ics'],
            'known_params'       : content['known'],
            'decomposition'      : content['decomp']
            }
        
    return data, content

###############################################################################
#Export as .py
###############################################################################    
def write_to_file(variables, equations, measured_states, states, unknown_parameters, input_conditions, ics, decomposition=[], known_params={}, filename=''):
    text     = '\n\n'.join([export_model(variables, equations),
                            export_args(measured_states, states, unknown_parameters, input_conditions, ics, decomposition, known_params),
                            export_run()])
    if len(filename) > 0:
        with open(filename, 'w') as file:
            file.write(text)
    return text

def export_model(variables, equations):
    global mod_prefix 
    
    header  = '###############################################################################\n#Model\n###############################################################################'
    mod     = mod_prefix + '.strike_goldd_simplified' if mod_prefix else 'strike_goldd_simplified'
    imports = 'from sympy import Matrix, Float, Symbol\nfrom ' + mod + ' import*'
    content = ''
    eq      = ''
    storage = 'variables = {' + ', '.join(["'" + v + "': " + v for group in variables for v in group]) + '}'
    
    ###############################################################################
    #Variables
    ###############################################################################   
    for line in variables:
        if len(line) > 1:
            lhs = ', '.join([x for x in line])
            rhs = '[Symbol(' + 'x' + ') for x in [' + ', '.join([x for x in quote(line)]) + ']]'
            content += lhs + ' = ' + rhs + '\n'
        elif len(line) == 1:
            lhs = line[0]
            rhs =  "Symbol('" + lhs + "')"
            content += lhs + ' = ' + rhs + '\n'
        else:
            continue

    ###############################################################################
    #Equations
    ############################################################################### 
    eq = '\n'.join([line for line in equations]) 

    result = '\n\n'.join([imports, header, content, eq, storage])
    return result

def export_args(measured_states, states, unknown_parameters, input_conditions, ics, decomposition, known_params):
    result  = '###############################################################################\n#Specify Input for IA\n###############################################################################\n'
    result += '#User makes changes here\n'
    
    def make_matrix(name, lst):
        return name + ' = Matrix([' + ', '.join(lst) + '])\n' 
    
    diff    = get_differential(states)
    result += make_matrix('measured_states   ', measured_states)
    result += make_matrix('states            ', states)
    result += make_matrix('unknown_parameters', unknown_parameters)
    result += make_matrix('diff              ', diff)
    result += get_input_conditions(input_conditions)
    result += get_ics(ics)
    result += get_decomposition(decomposition)
    
    ###############################################################################
    #Substitutions
    ###############################################################################
    if known_params:
        subs = get_substitution(known_params, unknown_parameters)
        subs += 'diff = diff.subs(known_parameters.items())'
        
        result += '\n\n' + subs
    
    return result

def get_differential(states):
    return ['d' + state for state in states]

def get_ics(ics):
    temp    = []
    for state in ics:
        value = str(state) + ': Float(' + str(ics[state]) + ', 3)'
        temp.append(value)

    return 'init_conditions    = {' + ', '.join(temp) + '}\n'

def get_input_conditions(input_conditions):
    temp    = []
    for state in input_conditions:
        value = str(state) + ': ' + str(input_conditions[state])
        temp.append(value)
           
    return 'input_conditions   = {' + ', '.join(temp) + '}\n'

def get_decomposition(decomposition):
    d = 'decomposition      = ['
    j = ',\n' + ' '* (len(d))
    c = '\n' + ' '* (len(d)) + ']\n'
    return d + j.join([ '[' + ', '.join(group) + ']' for group in decomposition]) + c

def get_substitution(known_params, unknown_parameters):
    temp    = []
    for param in known_params:
        if param in unknown_parameters:
            continue
        
        value = str(param) + ' : ' + str(known_params[param])
        temp.append(value)

    return 'known_parameters = {' + ', '.join(temp) + '}\n'
    
def export_run():
    
    s = '''
    ###############################################################################
    #Call Strike-Goldd
    ###############################################################################
    def run_strike_goldd():
        start_time = time()
        x_aug_dict = strike_goldd(measured_states, 
                                  states, 
                                  unknown_parameters, 
                                  input_conditions, 
                                  diff, 
                                  init_conditions, 
                                  decomposition)
        print("Total time: ",time()-start_time)
        print(x_aug_dict)
        return x_aug_dict
    
    if __name__ == '__main__':
        pass
    '''
    return dedent(s)

###############################################################################
#Permutational IA
###############################################################################
def permutational_IA(data):
    '''
    Takes in an imported model and returns a DataFrame of IA results for all possible combinations of observed variables/known parameters.
    '''
    param_perms  = permute_list(data['params'])
    output_perms = permute_list(data['outputs'])

    template = {**{key+'_Known':1 for key in data['params']},
                **{key+'_Known':0 for key in data['states']},
                **{key+'_Identifiable':1 for key in data['params']},
                **{key+'_Identifiable':0 for key in data['states']}
                }

    first       = True
    result_list = []

    for i in range(len(param_perms)):
        for ii in range(len(output_perms)):

            unknown_params   = param_perms[i]
            observed_outputs = output_perms[ii]
            print('unknown_params', unknown_params)
            print('observed_outputs', observed_outputs)

            IA_result = template.copy()
            for key in unknown_params:
                IA_result[key+'_Known'] = 0
            for key in observed_outputs:
                IA_result[key+'_Known'] = 1
            
            write_to_file(data, 'dummy.py', params=unknown_params, outputs=observed_outputs)

            if first:
                d     = importlib.import_module('dummy')
                first = False
            else:
                importlib.reload(d)
                
            try:
                result = d.runIA()
                
            except ZeroDivisionError:
                print('Skipping due to ZeroDivisionError')
                print('############################################')
                for key in IA_result:
                    IA_result[key] = -1
            else:
                print('############################################')
                for key in result:
                    IA_result[str(key) + '_Identifiable'] = int(result[key])

            
                result_list.append(list(IA_result.values()))

    result_dataframe = DataFrame(data=result_list, columns=list(template.keys()))
    
    return result_dataframe

def permute_list(x):
    '''
    First return value is all the combinations and permutations of the list x.
    '''
    result = []
    perms  = permute_true_false(len(x))
    for p in perms:
        temp = tuple([x[i] for i in range(len(p)) if p[i]])
        if temp:
            result.append(temp)

    return result
    
def permute_true_false(n):
    result = list(product([0, 1], repeat=n))
    
    return result

###############################################################################
#Interconversion with ConfigParser
###############################################################################
def txt_to_configparser(data, filename):
    try:
        configparser = importlib.import_module('configparser')
    except:
        pass
    config       = configparser.ConfigParser()

    for key in data:
        config[key] = {key: data[key]}
        
    with open(filename, 'w') as configfile:
        config.write(configfile)

s = '''

#parameters
vm,
synm1, synp1, 
degm, degp

#inputs
u = 1

#measured
Inde,
m1, p1

#known
degm  = 0.15,
synm1 = 1e-6,
synp1 = 0.01

#equations
dInde = -vm*Inde + u
dIndi =  vm*Inde
dm1 = synm1*Indi - degm*m1
dp1 = synp1*m1 - degp*p1

#states
Inde = 1,
Indi = 1,
m1   = 1, 
p1   = 1

#decomp
[Inde, Indi],
[m1, p1],
[Inde, Indi, m1, p1]
'''

if __name__ == '__main__':
    from   io import StringIO
    
    data, content = load_model(StringIO(s))
    
    model_name = 'inducible_with_input'
    filename    = model_name + '.py'
    
    code = write_to_file(**data, filename= model_name + '.py')    
    mod  = importlib.import_module( model_name)
    
    result = mod.run_strike_goldd()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

p = setuptools.find_packages()
print(p)

setuptools.setup(
    name                          = "BMSS", 
    version                       = "2.0.9",
    author                        = "Russell Ngo",
    author_email                  = "biernjk@gmail.com",
    description                   = "A package for dyanmic model analysis.",
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    url                           = "https://github.com/EngBioNUS/BMSS2",
    packages                      = setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.7',
    include_package_data = True,
    package_dir          = {'BMSS': 'BMSS'},
    package_data         = {'BMSS': ['models/*.db',
                                     'models/*.txt',
                                     'models/*.ini',
                                     'stylelib/*.mplstyle']}
)



import configparser
###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from .                   import model_handler    as mh 
    from .                   import settings_handler as sh
    from .setup_sim          import (compile_models, setup_helper,
                                     string_to_dict, string_to_dict_array, 
                                     string_to_list_string, eval_init_string, 
                                     eval_params_string, eval_tspan_string, 
                                     dict_template, list_template, t_template,
                                     is_analysis_settings)

except:
    import model_handler    as     mh
    import settings_handler as     sh
    from   setup_sim        import (compile_models, setup_helper,
                                    string_to_dict, string_to_dict_array, 
                                    string_to_list_string, eval_init_string, 
                                    eval_params_string, eval_tspan_string, 
                                    dict_template, list_template,
                                    is_analysis_settings)
    
###############################################################################
#Interfacing with Configparser
###############################################################################
def from_config(filename):
    '''Opens a config file and reads the fields/subfields required for setting up 
    the analysis while ignoring the irrelavant ones. Returns a dictionary of the 
    collected information.
    
    :param filename: Name of file to read.
    :type filename: str
    '''
    config             = configparser.RawConfigParser()
    config.optionxform = lambda option: option
    config_data               = {}
    with open(filename, 'r') as file:
        config.read_file(file)
    
    n = 1
    for section in config.sections():
        if not is_analysis_settings(config, section):
            continue
        
        init             = config[section]['init']
        params           = config[section]['parameter_values']
        tspan            = config[section]['tspan']
        solver_args      = config[section].get('solver_args')
        fixed_parameters = config[section].get('fixed_parameters')
        parameter_bounds = config[section].get('parameter_bounds')
        units            = config[section].get('units')
        
        init             = eval_init_string(init)
        tspan            = eval_tspan_string(tspan)
        params           = eval_params_string(params)
        solver_args      = string_to_dict(solver_args)             if solver_args      else {}
        fixed_parameters = string_to_list_string(fixed_parameters) if fixed_parameters else []
        parameter_bounds = string_to_dict_array(parameter_bounds)  if parameter_bounds else {}
        units            = string_to_dict(units)                   if units            else {}
        
        config_data[n] = {'system_type'      : section,     'init'             : init,      
                          'params'           : params,      'tspan'            : tspan,            
                          'solver_args'      : solver_args, 'fixed_parameters' : fixed_parameters, 
                          'parameter_bounds' : parameter_bounds,
                          'units'            : units
                          }
        
        n += 1
        
    return config_data

###############################################################################
#Main Set Up
###############################################################################    
def get_sensitivity_args(filename, user_core_models={}):
    '''Reads the config file and adds combines it with core_model data. If you are
    using a core model that is not in the database, you must provide the core_model
    using the core_models argument where the key is the system_type. Returns a 
    dictionary of keyword arguments and the config_data.
    
    :param filename: The name of the file to read
    :type filename: str
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    config_data, core_models = setup_helper(filename, from_config, user_core_models)
    models, params = compile_models(core_models, config_data)
    
    bounds           = {}
    fixed_parameters = []
    
    for key in config_data:
        models[key]['init']                    = config_data[key]['init']
        models[key]['tspan']                   = config_data[key]['tspan']
        models[key]['int_args']['solver_args'] = config_data[key]['solver_args']
        
        temp      = {param + '_' + str(key): config_data[key]['parameter_bounds'][param] for param in config_data[key]['parameter_bounds']}
        bounds    = {**bounds, **temp} 
        
        fixed_parameters   += [param + '_' + str(key) for param in config_data[key]['fixed_parameters']]
    
    sensitivity_args =  {'models'           : models, 
                         'params'           : params, 
                         'parameter_bounds' : bounds,
                         'fixed_parameters' : fixed_parameters,
                         'analysis_type'    : 'sobol',
                         'N'                : 1000 
                         }
    
    return sensitivity_args, config_data

###############################################################################
#Template Generation
###############################################################################    
def make_settings_template(system_types_settings_names, filename='', user_core_models={}):
    '''Writes settings to a config file. If you are using a core_model that is 
    not in the database, you must provide the core_model using the core_models 
    argument where the key is the system_type.Returns the code as a string.
    
    :param system_types_settings_names: Pairs of tuples containing (system_type, settings_name)
    :type system_types_settings_names: list or tuple
    :param filename: The name of the file to write to
    :type filename: str, optional
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    result = ''
    
    system_types_settings_names1 = [system_types_settings_names] if type(system_types_settings_names) == str else system_types_settings_names
    
    for pair in system_types_settings_names1:
        try:
            system_type, settings_name = pair
        except:
            system_type, settings_name = pair, '__default__'

        core_model     = user_core_models[system_type] if system_type in user_core_models else mh.quick_search(system_type)
        parameters     = core_model['parameters'] + core_model['inputs']
        states         = core_model['states']
        section_header = '[' + core_model['system_type'] + ']'
        
        settings = {'system_type'     : system_type, 
                    'settings_name'   : settings_name,
                    'parameters'      : {},
                    'units'           : {},
                    'init'            : {},  
                    'priors'          : {}, 
                    'parameter_bounds': {},
                    'tspan'            : [],
                    'fixed_parameters' : [],
                    'solver_args'      : {'rtol'   : 1.49012e-8,
                                          'atol'   : 1.49012e-8,
                                          'tcrit'  : [],
                                          'h0'     : 0.0,
                                          'hmax'   : 0.0,
                                          'hmin'   : 0.0,
                                          'mxstep' : 0
                                          },
                    }
        
        if settings_name:
            settings = sh.quick_search(system_type, settings_name, skip_constructor=True)
        
        longest          = len(max(parameters,               key=len))    
        longest_         = len(max(settings['solver_args'],  key=len))  
        solver_args      = settings['solver_args'].keys()
        init_values      = dict_template('init',             states,     longest, settings['init'])
        param_values     = dict_template('parameter_values', parameters, longest, settings['parameters'])
        parameter_bounds = dict_template('parameter_bounds', parameters, longest, settings['parameters'])
        units_values     = dict_template('units',            parameters, longest, settings['units'], default='')
        tspan            = ['[' + ', '.join(['{:.2f}'.format(x) for x in segment]) + ']' for segment in settings['tspan']]
        tspan            = t_template('tspan',            '[' + ', '.join(tspan) + ']')
        fixed_parameters = list_template('fixed_parameters', settings['fixed_parameters'])
        solver_args      = dict_template('solver_args',      solver_args, longest_, settings['solver_args'])
        
        
        model_id         = '#id = ' + str(core_model['id'])
        model_equations  = '#equations = \n' + '\n'.join(['#\t' + line for line in core_model['equations'] ])
        section_header   = '\n'.join([section_header, model_id, model_equations, ''])
        
        result += '\n\n'.join([section_header, init_values, param_values, units_values, tspan, parameter_bounds, fixed_parameters, solver_args])
        
    if filename:
        with open(filename, 'w') as file:
            file.write(result)
    return result





import configparser
from   sympy    import Matrix, Float, Symbol
from   textwrap import dedent

###############################################################################
#Non-Standard Imports
###############################################################################


try:
    from .                   import model_handler    as mh 
    from .                   import settings_handler as sh
    from .                   import ia_model_import  as im
    from .setup_sim          import (compile_models, setup_helper, 
                                     string_to_dict, string_to_dict_array, 
                                     string_to_list_string, eval_init_string, 
                                     eval_params_string, eval_tspan_string, 
                                     dict_template, list_template,
                                     split_at_top_level,
                                     is_analysis_settings)

except:
    import model_handler    as     mh
    import settings_handler as     sh
    import ia_model_import  as     im
    from   setup_sim        import (compile_models, setup_helper, 
                                     string_to_dict, string_to_dict_array, 
                                     string_to_list_string, eval_init_string, 
                                     eval_params_string, eval_tspan_string, 
                                     dict_template, list_template, t_template,
                                     split_at_top_level,
                                     is_analysis_settings)

###############################################################################
#Interfacing with ConfigParser
###############################################################################  
def from_config(filename):
    '''Opens a config file and reads the fields/subfields required for setting up 
    the analysis while ignoring the irrelavant ones. Returns a dictionary of the 
    collected information.
    
    :param filename: Name of file to read.
    :type filename: str
    '''
    config             = configparser.RawConfigParser()
    config.optionxform = lambda option: option
    config_data               = {}
    with open(filename, 'r') as file:
        config.read_file(file)
    
    n = 1

    for section in config.sections():
        if not is_analysis_settings(config, section):
            continue
        
        parameter_values = config[section].get('parameter_values')
        measured_states  = config[section].get('measured_states')
        input_conditions = config[section].get('input_conditions')
        decomposition    = config[section].get('decomposition')
        init             = config[section].get('init')
        units            = config[section].get('units')
        
        measured_states  = string_to_list_string(measured_states)  if measured_states  else []
        input_conditions = string_to_dict(input_conditions)        if input_conditions else {}
        parameter_values = string_to_dict(parameter_values)        if parameter_values else {} 
        units            = string_to_dict(units)                   if units            else {}
        init             = eval_init_string(init)                  if init             else {}
        
        if decomposition:
            temp = ''.join([s for s in decomposition if s in ['[', ']']])
            if '[[' in temp:
                decomposition = decomposition.strip()
                decomposition = decomposition[1: len(decomposition)-1]
        else:
            decomposition = []
            
        decomposition    = [s.strip() for s in split_at_top_level(decomposition)]
        decomposition    = [[v.strip() for v in s.replace(']', '').replace('[', '').split(',')] for s in decomposition]
        
        for key in init:
            try: 
                next(iter(init[key]))
                if len(init[key]) != 1:
                    raise Exception('Error in init. Only one value allowed.')
            except:
                pass
            
        config_data[n] = {'system_type'      : section, 
                          'parameter_values' : parameter_values,
                          'init'             : init,
                          'input_conditions' : input_conditions,
                          'measured_states'  : measured_states, 
                          'decomposition'    : decomposition,
                          'units'            : units
                          }
        n += 1
        
    return config_data

###############################################################################
#Main Set Up
###############################################################################    
def get_strike_goldd_args(filename, user_core_models={}, write_file=False):
    '''Reads the config file and adds combines it with core_model data. If you are
    using a core model that is not in the database, you must provide the core_model
    using the core_models argument where the key is the system_type. Returns a 
    dictionary of keyword arguments and the config_data.
    
    :param filename: The name of the file to read
    :type filename: str
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    config_data, core_models = setup_helper(filename, from_config, user_core_models)
    
    all_variables = {}
    sg_args       = {}
    
    for model_num, core_model in zip(config_data, core_models):
        var     = [core_model[key] for key in ['states', 'parameters', 'inputs']]
        eq      = core_model['equations']
        h       = config_data[model_num].get('measured_states')
        x       = core_model['states']
        p       = [v for v in core_model['parameters'] if v not in config_data[model_num].get('parameter_values', {}) ]
        u       = config_data[model_num].get('input_conditions', {})
        ics     = config_data[model_num].get('init')
        ics     = dict(zip(x, ics[next(iter(ics))]))
        decomp  = config_data[model_num].get('decomposition')
        values  = config_data[model_num].get('parameter_values', {})

        if write_file: 
            outfile = 'sg_' + core_model['system_type'].replace(', ', '_') + '.py'
        else:
            outfile = ''

        mapping = [v for key in ['states', 'parameters', 'inputs'] for v in core_model[key]]
        mapping = {Symbol(v) : Symbol(v + '_' + str(model_num)) for v in mapping}
        code    = im.write_to_file(var, eq, h, x, p, u, ics, decomp, values, outfile)
        data    = {}
        temp    = []

        save_variables = '''
        data["h"]      = measured_states
        data["x"]      = states
        data["p"]      = unknown_parameters
        data["f"]      = diff
        data["u"]      = input_conditions
        data["ics"]    = init_conditions
        data["decomp"] = decomposition
        
        temp.append(variables)
        '''
        exec(code)
        exec(dedent(save_variables))

        items = mapping.items()
        for key in data :
            if key in ['h', 'x', 'p', 'f']:
                data[key] = data[key].subs(items)
            elif key in ['u', 'ics']:
                data[key] = {k.subs(items): data[key][k] for k in data[key]}
            else:
                data[key] = [[v.subs(items) for v in group] for group in data[key]]
        
        temp = {key: temp[0][key].subs(items) for key in temp[0]}
        
        sg_args[model_num]       = data
        all_variables[model_num] = temp
            
    return sg_args, config_data, all_variables

###############################################################################
#Template Generation
###############################################################################    
def make_settings_template(system_types_settings_names, filename='', user_core_models={}):
    '''Writes settings to a config file. If you are using a core_model that is 
    not in the database, you must provide the core_model using the core_models 
    argument where the key is the system_type.Returns the code as a string.
    
    :param system_types_settings_names: Pairs of tuples containing (system_type, settings_name)
    :type system_types_settings_names: list or tuple
    :param filename: The name of the file to write to
    :type filename: str, optional
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    result = ''
    
    system_types_settings_names1 = [system_types_settings_names] if type(system_types_settings_names) == str else system_types_settings_names
    
    for pair in system_types_settings_names1:
        try:
            system_type, settings_name = pair
        except:
            system_type, settings_name = pair, '__default__'

        core_model     = user_core_models[system_type] if system_type in user_core_models else mh.quick_search(system_type)
        parameters     = core_model['parameters']
        inputs         = core_model['inputs']
        states         = core_model['states']
        section_header = '[' + core_model['system_type'] + ']'
        
        settings = {'system_type'     : system_type, 
                    'settings_name'   : settings_name,
                    'parameters'      : {},
                    'units'           : {},
                    'init'            : {},  
                    'priors'          : {}, 
                    'parameter_bounds': {},
                    'tspan'            : [],
                    'fixed_parameters' : [],
                    'solver_args'      : {'rtol'   : 1.49012e-8,
                                          'atol'   : 1.49012e-8,
                                          'tcrit'  : [],
                                          'h0'     : 0.0,
                                          'hmax'   : 0.0,
                                          'hmin'   : 0.0,
                                          'mxstep' : 0
                                          },
                    }
        
        if settings_name:
            settings = sh.quick_search(system_type, settings_name, skip_constructor=True)
        
        longest            = len(max(parameters,               key=len))    
        longest_           = len(max(inputs,                   key=len))  if inputs else None
        init_values        = dict_template('init',             states,     longest,  settings['init'],       default = 0)
        param_values       = dict_template('parameter_values', parameters, longest,  settings['parameters'], default = 1)
        input_conditions   = dict_template('input_conditions', inputs,     longest_, dict(zip(inputs, [0]*len(inputs)))) if inputs else ''
        fixed_parameters   = list_template('fixed_parameters', settings['fixed_parameters'])
        measured_states    = list_template('measured_states', states)
        
        model_id         = '#id = ' + str(core_model['id'])
        model_equations  = '#equations = \n' + '\n'.join(['#\t' + line for line in core_model['equations'] ])
        section_header   = '\n'.join([section_header, model_id, model_equations])
        
        result += '\n\n'.join([section_header, param_values, init_values, fixed_parameters, measured_states, input_conditions])
        
    if filename:
        with open(filename, 'w') as file:
            file.write(result)
    return result
